@@ -10,17 +10,13 @@ from opt_einsum import contract
 import psi4
 psi4.core.be_quiet()
 
-try:
-    from rich import print
-except:
-    pass
-
 try:
     from pylibxc import LibXCFunctional as Functional
 except:
     pass
 
 from .cubeprop import Cubeprop
+from .basis_set_artifact_correction import basis_set_correction, invert_kohn_sham_equations
 
 class Grider(Cubeprop):
 
@@ -60,6 +56,7 @@ class Grider(Cubeprop):
         nblocks = int(np.floor(npoints/max_points))
         blocks = []
 
+        max_functions = 0
         #Run through full blocks
         idx = 0
         for nb in range(nblocks):
@@ -73,6 +70,8 @@ class Grider(Cubeprop):
 
             blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))
             idx += max_points
+            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
+                                          else len(blocks[-1].functions_local_to_global())
 
         #Run through remaining points
         if idx < npoints:
@@ -85,13 +84,16 @@ class Grider(Cubeprop):
                 w = psi4.core.Vector.from_array(np.zeros_like(grid[2][idx:]))  # When w is not necessary and not given
             blocks.append(psi4.core.BlockOPoints(x, y, z, w, extens))
 
-        max_functions = 0 if 0 > len(blocks[-1].functions_local_to_global()) \
-                                      else len(blocks[-1].functions_local_to_global())
+            max_functions = max_functions if max_functions > len(blocks[-1].functions_local_to_global()) \
+                                          else len(blocks[-1].functions_local_to_global())
 
+        zero_matrix = psi4.core.Matrix(basis.nbf(), basis.nbf())
         if self.ref == 1:
             point_func = psi4.core.RKSFunctions(basis, max_points, max_functions)
+            point_func.set_pointers(zero_matrix)
         else:
             point_func = psi4.core.UKSFunctions(basis, max_points, max_functions)
+            point_func.set_pointers(zero_matrix, zero_matrix)
 
         return blocks, npoints, point_func
 
@@ -120,13 +122,14 @@ class Grider(Cubeprop):
 
         return grid, shape
 
-    def generate_dft_grid(self, wfn):
+    def generate_dft_grid(self, vpot):
         """
         Extracts DFT spherical grid and weights from wfn object
 
         Parameters
         ----------
-        wfn: psi4.core.{RKS, UKS}
+        vpot: psi4.core.VBase
+            Vpot object with dft grid data
 
         Returns
         -------
@@ -136,16 +139,11 @@ class Grider(Cubeprop):
         
         """
 
-        try:
-            vpot = wfn.V_potential()
-        except:
-            raise ValueError("Wfn object does not contain a Vpotential object. Try with one obtained from a DFT calculation")
-
         nblocks = vpot.nblocks()
         blocks = [vpot.get_block(i) for i in range(nblocks)]
         npoints = vpot.grid().npoints()
 
-        dft_grid = np.empty(4, npoints)
+        dft_grid = np.empty((4, npoints))
 
         offset = 0
         for i_block in blocks:
@@ -159,7 +157,6 @@ class Grider(Cubeprop):
 
         return dft_grid
 
-
     #Quantities on Grid
     def on_grid_ao(self, coeff, grid=None, basis=None, vpot=None):
         """
@@ -170,7 +167,7 @@ class Grider(Cubeprop):
         ----------
         coeff: np.ndarray
             Vector/Matrix of quantity on ao basis. Shape: {(num_ao_basis, ), (num_ao_basis, num_ao_basis)}
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed.
         basis: psi4.core.BasisSet, optional
             The basis set. If not given it will use target wfn.basisset().
@@ -188,10 +185,13 @@ class Grider(Cubeprop):
 
 
         if grid is not None:
-            if grid.shape[0] != 3 and grid.shape[0] != 4:
-                raise ValueError("The shape of grid should be (3, npoints) "
-                                 "or (4, npoints) but got (%i, %i)"%(grid.shape[0], grid.shape[1]))
-            blocks, npoints, points_function = self.grid_to_blocks(grid, basis=basis)
+            if type(grid) is np.ndarray:
+                if grid.shape[0] != 3 and grid.shape[0] != 4:
+                    raise ValueError("The shape of grid should be (3, npoints) "
+                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
+                blocks, npoints, points_function = self.grid_to_blocks(grid, basis=basis)
+            else:
+                blocks, npoints, points_function = grid
         elif grid is None and vpot is not None:
             nblocks = vpot.nblocks()
             blocks = [vpot.get_block(i) for i in range(nblocks)]
@@ -200,14 +200,7 @@ class Grider(Cubeprop):
         else:
             raise ValueError("A grid or a V_potential (DFT grid) must be given.")
 
-        if self.ref == 1:
-            psi4_coeff = psi4.core.Matrix.from_array(coeff[:,None])
-            points_function.set_pointers(psi4_coeff)
-        elif self.ref == 2:
-            coeff_0 = coeff[:,None] if coeff.ndim == 1 else coeff.copy()
-            psi4_coeff = psi4.core.Matrix.from_array(coeff_0)
-            points_function.set_pointers(psi4_coeff, psi4_coeff)
-        coeff_r = np.empty((npoints)) 
+        coeff_r = np.empty((npoints))
 
         offset = 0
         for i_block in blocks:
@@ -215,12 +208,14 @@ class Grider(Cubeprop):
             b_points = i_block.npoints()
             offset += b_points
             lpos = np.array(i_block.functions_local_to_global())
+            if len(lpos)==0:
+                continue
             phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]
 
             if coeff.ndim == 1:
                 l_mat = coeff[(lpos[:])]
                 coeff_r[offset - b_points : offset] = contract('pm,m->p', phi, l_mat)
-            elif coeff.ndim == 2: 
+            elif coeff.ndim == 2:
                 l_mat = coeff[(lpos[:, None], lpos)]
                 coeff_r[offset - b_points : offset] = contract('pm,mn,pn->p', phi, l_mat, phi)
 
@@ -237,7 +232,7 @@ class Grider(Cubeprop):
         ----------
         Da, Db: np.ndarray
             Alpha, Beta densities. Shape: (num_ao_basis, num_ao_basis)
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed.
         vpot: psi4.core.VBase
             Vpotential object with info about grid.
@@ -245,7 +240,7 @@ class Grider(Cubeprop):
 
         Returns
         -------
-        density: np.ndarray Shape: (npoints, )
+        density: np.ndarray Shape: (ref, npoints)
             Density on the given grid. 
         """
 
@@ -260,10 +255,13 @@ class Grider(Cubeprop):
             raise ValueError("Db is required for an unrestricted system")
 
         if grid is not None:
-            if grid.shape[0] != 3 and grid.shape[0] != 4:
-                raise ValueError("The shape of grid should be (3, npoints) "
-                                 "or (4, npoints) but got (%i, %i)"%(grid.shape[0], grid.shape[1]))
-            blocks, npoints, points_function = self.grid_to_blocks(grid)
+            if type(grid) is np.ndarray:
+                if grid.shape[0] != 3 and grid.shape[0] != 4:
+                    raise ValueError("The shape of grid should be (3, npoints) "
+                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
         elif grid is None and vpot is not None:
             nblocks = vpot.nblocks()
             blocks = [vpot.get_block(i) for i in range(nblocks)]
@@ -272,27 +270,32 @@ class Grider(Cubeprop):
         else:
             raise ValueError("A grid or a V_potential (DFT grid) must be given.")
 
-        density   = np.empty((npoints, self.ref))
+        
         if self.ref == 1:
             points_function.set_pointers(Da)
             rho_a = points_function.point_values()["RHO_A"]
+            density   = np.empty((npoints))
         if self.ref == 2:
             points_function.set_pointers(Da, Db)
             rho_a = points_function.point_values()["RHO_A"]
             rho_b = points_function.point_values()["RHO_B"]
+            density   = np.empty((npoints, self.ref))
 
         offset = 0
         for i_block in blocks:
             points_function.compute_points(i_block)
             b_points = i_block.npoints()
             offset += b_points
-            density[offset - b_points : offset, 0] = 0.5 * rho_a.np[ :b_points]
-            if self.ref == 2:
-                density[offset - b_points : offset, 1] = 0.5 * rho_b.np[ :b_points]
+
+            if self.ref == 1:
+                density[offset - b_points : offset] = rho_a.np[ :b_points]
+            else:
+                density[offset - b_points : offset, 0] = rho_a.np[ :b_points]
+                density[offset - b_points : offset, 1] = rho_b.np[ :b_points]
 
         return density
 
-    def on_grid_orbitals(self, grid=None, Ca=None, Cb=None, vpot=None):
+    def on_grid_orbitals(self, Ca=None, Cb=None, grid=None, vpot=None):
         """
         Generates orbitals given grid
 
@@ -300,7 +303,7 @@ class Grider(Cubeprop):
         ----------
         Ca, Cb: np.ndarray
             Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed
         vpot: psi4.core.VBase
             Vpotential object with info about grid.
@@ -326,10 +329,13 @@ class Grider(Cubeprop):
             raise ValueError("Db is required for an unrestricted system")
 
         if grid is not None:
-            if grid.shape[0] != 3 and grid.shape[0] != 4:
-                raise ValueError("The shape of grid should be (3, npoints) "
-                                 "or (4, npoints) but got (%i, %i)"%(grid.shape[0], grid.shape[1]))
-            blocks, npoints, points_function = self.grid_to_blocks(grid)
+            if type(grid) is np.ndarray:
+                if grid.shape[0] != 3 and grid.shape[0] != 4:
+                    raise ValueError("The shape of grid should be (3, npoints) "
+                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
         elif grid is None and vpot is not None:
             nblocks = vpot.nblocks()
             blocks = [vpot.get_block(i) for i in range(nblocks)]
@@ -338,15 +344,15 @@ class Grider(Cubeprop):
         else:
             raise ValueError("A grid or a V_potential (DFT grid) must be given.")
 
-        orbitals_r = [np.empty((npoints, self.ref)) for i_orb in range(self.nbf)]
-
         if self.ref == 1:
+            orbitals_r = [np.empty((npoints)) for i_orb in range(self.nbf)]
             points_function.set_pointers(Ca)
+            Ca_np = Ca.np
         if self.ref == 2:
+            orbitals_r = [np.empty((npoints, 2)) for i_orb in range(self.nbf)]
             points_function.set_pointers(Ca, Cb)
-
-        Ca_np = Ca.np
-        Cb_np = Cb.np
+            Ca_np = Ca.np
+            Cb_np = Cb.np
 
         offset = 0
         for i_block in blocks:
@@ -354,41 +360,65 @@ class Grider(Cubeprop):
             b_points = i_block.npoints()
             offset += b_points
             lpos = np.array(i_block.functions_local_to_global())
+            if len(lpos)==0:
+                continue
             phi = np.array(points_function.basis_values()["PHI"])[:b_points, :lpos.shape[0]]
 
             for i_orb in range(self.nbf):
                 Ca_local = Ca_np[lpos, i_orb]
-                Cb_local = Cb_np[lpos, i_orb]
-                orbitals_r[i_orb][offset - b_points : offset,0] = contract('m, pm -> p', Ca_local, phi)
-                orbitals_r[i_orb][offset - b_points : offset,1] = contract('m, pm -> p', Cb_local, phi)
+                if self.ref == 1:
+                    orbitals_r[i_orb][offset - b_points : offset] = contract('m, pm -> p', Ca_local, phi)
+                else:
+                    Cb_local = Cb_np[lpos, i_orb]
+                    orbitals_r[i_orb][offset - b_points : offset,0] = contract('m, pm -> p', Ca_local, phi)
+                    orbitals_r[i_orb][offset - b_points : offset,1] = contract('m, pm -> p', Cb_local, phi)
+
+
+
 
         return orbitals_r
 
-    def on_grid_esp(self, grid=None, vpot=None):
+    def on_grid_esp(self, Da=None, Db=None, grid=None, vpot=None, wfn=None):
 
         """
         Generates EXTERNAL/ESP/HARTREE and Fermi Amaldi Potential on given grid
 
         Parameters
         ----------
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+        Da,Db: np.ndarray, opt, shape (nbf, nbf)
+            The electron density in the denominator of Hartee potential. If None, the original density matrix
+            will be used.
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed.
         vpot: psi4.core.VBase
             Vpotential object with info about grid.
             Provides DFT spherical grid. Only comes to play if no grid is given. 
-
+        
         Returns
         -------
         vext, hartree, esp, v_fa: np.ndarray
             External, Hartree, ESP, and Fermi Amaldi potential on the given grid
             Shape: (npoints, )
         """
+        if Da is not None or Db is not None:
+            Da_temp = np.copy(self.wfn.Da().np)
+            Db_temp = np.copy(self.wfn.Db().np)
+            if Da is not None:
+                self.wfn.Da().np[:] = Da
+            if Db is not None:
+                self.wfn.Db().np[:] = Db
 
         nthreads = psi4.get_num_threads()
         psi4.set_num_threads(1)
-
+        
+        if wfn is None:
+            wfn = self.wfn
+        
         if grid is not None:
-            blocks, npoints, points_function = self.grid_to_blocks(grid)
+            if type(grid) is np.ndarray:
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
         elif grid is None and vpot is not None:
             nblocks = vpot.nblocks()
             blocks = [vpot.get_block(i) for i in range(nblocks)]
@@ -409,7 +439,7 @@ class Grider(Cubeprop):
         zs = [mol_dict["elez"][i] for i in indx]
         rs = [self.mol.geometry().np[i] for i in indx]
 
-        esp_wfn = psi4.core.ESPPropCalc(self.wfn)
+        esp_wfn = psi4.core.ESPPropCalc(wfn)
 
         #Loop Through blocks
         offset = 0
@@ -436,6 +466,10 @@ class Grider(Cubeprop):
         hartree = - 1.0 * (vext + esp)
         v_fa = (1 - 1.0 / (self.nalpha + self.nbeta)) * hartree
 
+        if Da is not None:
+            self.wfn.Da().np[:] = Da_temp
+        if Db is not None:
+            self.wfn.Db().np[:] = Db_temp
         psi4.set_num_threads(nthreads)
 
         return vext, hartree, v_fa, esp
@@ -452,7 +486,7 @@ class Grider(Cubeprop):
         func_id: int
             Functional ID associated with Density Functional Approximationl.
             Full list of functionals: https://www.tddft.org/programs/libxc/functionals/
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed.
         vpot: psi4.core.VBase
             Vpotential object with info about grid.
@@ -469,26 +503,22 @@ class Grider(Cubeprop):
         if func_id != 1:
             raise ValueError("Only LDA fucntionals are supported on the grid")
 
-        if  Da is None and Db is None and vpot is None:
-            density = self.on_grid_density(grid=grid)
-        elif Da is not None and Db is None and vpot is None:
-            density = self.on_grid_density(grid=grid, Da=Da)
-        elif Da is not None and Db is not None and vpot is None:
-            density = self.on_grid_density(grid=grid, Da=Da, Db=Db)
-        ###
-        elif Da is None and Db is None and vpot is not None:
-            density = self.on_grid_density(vpot=vpot)        
-        elif Da is not None and Db is None and vpot is not None:
-            density = self.on_grid_density(Da=Da, vpot=vpot)
-        elif Da is not None and Db is not None and vpot is not None:
-            density = self.on_grid_density(Da=Da, Db=Db, vpot=vpot)
+        if Da is None:
+            Da = self.nt[0]
+        if Db is None:
+            Db = self.nt[0]
 
         if grid is not None:
-            blocks, npoints, points_function = self.grid_to_blocks(grid)
+            if type(grid) is np.ndarray:
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
+            density = self.on_grid_density(Da=Da, Db=Db, grid=grid)
         elif grid is None and vpot is not None:
             nblocks = vpot.nblocks()
             blocks = [vpot.get_block(i) for i in range(nblocks)]
             npoints = vpot.grid().npoints()
+            density = self.on_grid_density(Da=Da, Db=Db, vpot=vpot)
         else:
             raise ValueError("A grid or a V_potential (DFT grid) must be given.")
 
@@ -498,57 +528,188 @@ class Grider(Cubeprop):
         for i_block in blocks:
             b_points = i_block.npoints()
             offset += b_points
-            ingredients["rho"] = density[offset - b_points : offset, :]
+            if self.ref == 1:
+                ingredients["rho"] = density[offset - b_points : offset]
+            else:
+                ingredients["rho"] = density[offset - b_points : offset, :]
 
             if self.ref == 1:
                 functional = Functional(1, 1)
             else:
                 functional = Functional(1, 2) 
             xc_dictionary = functional.compute(ingredients)
-            vxc[offset - b_points : offset, :] = np.squeeze(xc_dictionary['vrho'])
+            vxc[offset - b_points : offset, :] = xc_dictionary['vrho']
 
-        return vxc
+        return np.squeeze(vxc)
 
-    def on_grid_all(self, grid=None, show_progress=False):
+    def on_grid_lap_phi(self, 
+                       Ca=None, 
+                       Cb=None, 
+                       grid=None,
+                       vpot=None):
         """
-        Calls *all* 'on_grid_x' functions using densities and coefficients stored in Inverter object. 
+        Generates laplacian of molecular orbitals
 
         Parameters
         ----------
-        grid: np.ndarray Shape: (3, npoints) or (4, npoints)
+       Ca, Cb: np.ndarray
+            Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
             grid where density will be computed.
-        show_progress: bool 
-            If True. Prints confirmation of each component generated. 
+        vpot: psi4.core.VBase
+            Vpotential object with info about grid.
+            Provides DFT spherical grid. Only comes to play if no grid is given. 
 
         Returns
         -------
-        None:
-            *Stores all grid quantities on self.grid. 
-
+        lap_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
+            Laplacian of molecular orbitals on the grid
         """
 
-        if show_progress == True:
-            print("Generating Orbitals") 
-        density = self.on_grid_density(grid=grid)
-        if show_progress == True:
-            print("Generating Orbitals") 
-        orbitals = self.on_grid_orbitals(grid=grid)
-        if show_progress is True:
-            print("Generating Ext/Hartree/ESP") 
-        vext, hartree, esp, v_fa = self.on_grid_esp(wfn=self.wfn, grid=grid)
-        if show_progress is True:
-            print("Generating Vxc") 
-        vxc = self.on_grid_vxc(grid=grid)
+        if Ca is None and Cb is None:
+            Ca = psi4.core.Matrix.from_array(self.Ca)
+            Cb = psi4.core.Matrix.from_array(self.Cb)
+        else:
+            Ca = psi4.core.Matrix.from_array(Ca)
+            Cb = psi4.core.Matrix.from_array(Cb)
 
-        self.grid.density = density
-        self.grid.orbitals = orbitals
-        self.grid.vext = vext
-        self.grid.hartree = hartree
-        self.grid.esp = esp
-        self.grid.v_fa = v_fa
-        self.grid.vxc = vxc
+        if self.ref == 2 and Cb is None:
+            raise ValueError("Db is required for an unrestricted system")
 
-        return
+        if grid is not None:
+            if type(grid) is np.ndarray:
+                if grid.shape[0] != 3 and grid.shape[0] != 4:
+                    raise ValueError("The shape of grid should be (3, npoints) "
+                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
+        elif grid is None and vpot is not None:
+            nblocks = vpot.nblocks()
+            blocks = [vpot.get_block(i) for i in range(nblocks)]
+            npoints = vpot.grid().npoints()
+            points_function = vpot.properties()[0]
+        else:
+            raise ValueError("A grid or a V_potential (DFT grid) must be given.")
+
+        points_function.set_ansatz(2)
+
+        if self.ref == 1:
+            points_function.set_pointers(Ca)
+            lap_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
+        else:
+            points_function.set_pointers(Ca, Cb)
+            lap_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]
+
+        offset = 0
+        for i_block in blocks:
+            points_function.compute_points(i_block)
+            b_points = i_block.npoints()
+            offset += b_points
+            lpos = np.array(i_block.functions_local_to_global())
+            if len(lpos)==0:
+                continue
+            
+            #Obtain subset of phi_@@ matrices
+            lx = np.array(points_function.basis_values()["PHI_XX"])[:b_points, :lpos.shape[0]]
+            ly = np.array(points_function.basis_values()["PHI_YY"])[:b_points, :lpos.shape[0]]
+            lz = np.array(points_function.basis_values()["PHI_ZZ"])[:b_points, :lpos.shape[0]]
+
+            for i_orb in range(self.nbf):
+                Ca_local = Ca.np[lpos, i_orb][:,None]
+                
+                if self.ref ==1:
+                    lap_phi[i_orb][offset - b_points : offset] += ((lx + ly + lz) @ Ca_local)[:,0]
+                else:
+                    Cb_local = Cb.np[lpos, i_orb][:,None]
+                    lap_phi[i_orb][offset - b_points : offset, 0] += ((lx + ly + lz) @ Ca_local)[:,0]
+                    lap_phi[i_orb][offset - b_points : offset, 1] += ((lx + ly + lz) @ Cb_local)[:,0]
+
+        return lap_phi
+
+    def on_grid_grad_phi(self, 
+                       Ca=None, 
+                       Cb=None, 
+                       grid=None,
+                       vpot=None):
+        """
+        Generates laplacian of molecular orbitals
+
+        Parameters
+        ----------
+       Ca, Cb: np.ndarray
+            Alpha, Beta Orbital Coefficient Matrix. Shape: (num_ao_basis, num_ao_basis)
+        grid: np.ndarray Shape: (3, npoints) or (4, npoints) or tuple for block_handler (return of grid_to_blocks)
+            grid where density will be computed.
+        vpot: psi4.core.VBase
+            Vpotential object with info about grid.
+            Provides DFT spherical grid. Only comes to play if no grid is given. 
+
+        Returns
+        -------
+        grad_phi: List[np.ndarray]. Where array is of shape (npoints, ref)
+            Gradient of molecular orbitals on the grid
+        """
+
+        if Ca is None and Cb is None:
+            Ca = psi4.core.Matrix.from_array(self.Ca)
+            Cb = psi4.core.Matrix.from_array(self.Cb)
+        else:
+            Ca = psi4.core.Matrix.from_array(Ca)
+            Cb = psi4.core.Matrix.from_array(Cb)
+
+        if self.ref == 2 and Cb is None:
+            raise ValueError("Db is required for an unrestricted system")
+
+        if grid is not None:
+            if type(grid) is np.ndarray:
+                if grid.shape[0] != 3 and grid.shape[0] != 4:
+                    raise ValueError("The shape of grid should be (3, npoints) "
+                                     "or (4, npoints) but got (%i, %i)" % (grid.shape[0], grid.shape[1]))
+                blocks, npoints, points_function = self.grid_to_blocks(grid)
+            else:
+                blocks, npoints, points_function = grid
+        elif grid is None and vpot is not None:
+            nblocks = vpot.nblocks()
+            blocks = [vpot.get_block(i) for i in range(nblocks)]
+            npoints = vpot.grid().npoints()
+            points_function = vpot.properties()[0]
+        else:
+            raise ValueError("A grid or a V_potential (DFT grid) must be given.")
+
+        points_function.set_ansatz(2)
+
+        if self.ref == 1:
+            points_function.set_pointers(Ca)
+            grad_phi = [np.zeros((npoints)) for i_orb in range(self.nbf)]
+        else:
+            points_function.set_pointers(Ca, Cb)
+            grad_phi = [np.zeros((npoints, 2)) for i_orb in range(self.nbf)]
+
+        offset = 0
+        for i_block in blocks:
+            points_function.compute_points(i_block)
+            b_points = i_block.npoints()
+            offset += b_points
+            lpos = np.array(i_block.functions_local_to_global())
+            if len(lpos)==0:
+                continue
+            
+            #Obtain subset of phi_@ matrix
+            gx = np.array(points_function.basis_values()["PHI_X"])[:b_points, :lpos.shape[0]]
+            gy = np.array(points_function.basis_values()["PHI_Y"])[:b_points, :lpos.shape[0]]
+            gz = np.array(points_function.basis_values()["PHI_Z"])[:b_points, :lpos.shape[0]]
+
+            for i_orb in range(self.nbf):
+                Ca_local = Ca.np[lpos, i_orb][:,None]
+                if self.ref == 1:
+                    grad_phi[i_orb][offset - b_points : offset] += ((gx + gy + gz) @ Ca_local)[:,0]
+                if self.ref == 2:
+                    Cb_local = Cb.np[lpos, i_orb][:,None]
+                    grad_phi[i_orb][offset - b_points : offset, 0] += ((gx + gy + gz) @ Ca_local)[:,0]
+                    grad_phi[i_orb][offset - b_points : offset, 1] += ((gx + gy + gz) @ Cb_local)[:,0]
+
+        return grad_phi
 
     def dft_grid_to_fock(self, value, Vpot):
         """For value on DFT spherical grid, Fock matrix is returned.
@@ -578,7 +739,9 @@ class Grider(Cubeprop):
             points_func.compute_points(block)
             npoints = block.npoints()
             lpos = np.array(block.functions_local_to_global())
-
+            if len(lpos) == 0:
+                i += npoints
+                continue
             # Obtain the grid weight
             w = np.array(block.w())
 
@@ -594,9 +757,6 @@ class Grider(Cubeprop):
         assert i == value.shape[0], "Did not run through all the points. %i %i" %(i, value.shape[0])
         return VFock
 
-
-
-
-
-
-
+    #Miscellaneous
+    def get_basis_set_correction(self, grid):
+        return basis_set_correction(self, grid)    
